@@ -17,6 +17,8 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -27,6 +29,7 @@
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
+#include "intl.h"
 #include "util.h"
 #include "gtypes.h"
 #include "detect.h"
@@ -48,6 +51,13 @@
 #define PREVIEW_WIDTH (min(gdk_screen_width()-440,420))
 #define PREVIEW_HEIGHT (min(gdk_screen_height()-300,360))
 
+typedef enum
+{
+	PASTE_MOVING,
+	PASTE_COPYING,
+	PASTE_LINKING
+} PasteType;
+
 GtkWidget *entry;
 GtkWidget *imagelist;
 GtkWidget *preview_box;
@@ -60,7 +70,6 @@ static GdkPixmap *gtksee_pixmap;
 static GdkBitmap *gtksee_mask;
 static GdkPixmap *blank_pixmap;
 static GdkBitmap *blank_mask;
-static GdkImage *blank_image;
 static GdkCursor *watch_cursor;
 
 static gchar current_selected_dir[256];
@@ -71,29 +80,37 @@ static time_t list_last_modified = 0;
 static gchar root[64];
 static gint busy_level;
 
-void		set_busy_cursor		();
-void		set_normal_cursor	();
+static guint clipboard_size = 0;
+static gboolean clipboard_moving = FALSE;
+static char **file_clipboard = NULL;
+
+static void	set_busy_cursor		();
+static void	set_normal_cursor	();
 
 
-int		main			(int argc,
-					 char *argv[]);
-void		dir_selected		(GtkTree *dt,
+static void	dir_selected		(GtkTree *dt,
 					 GtkWidget *child);
-void		dir_selected_internal	(GtkTree *dt,
+static void	dir_selected_internal	(GtkTree *dt,
 					 GtkWidget *child);
-void		file_selected		(ImageList *il);
-void		file_selected_internal	(ImageList *il);
-void		file_unselected		(ImageList *il);
-void		dir_entry_enter		(GtkWidget *widget, GtkWidget *entry);
-void		refresh_status		();
-void		preview_size_changed	(GtkWidget *box);
-void		imagelist_size_changed	(GtkWidget *box);
-void		clear_preview_box	();
-void		refresh_list		();
-void		auto_refresh_list	();
-void		show_viewer		();
-void		print_help_message	();
-void		print_error_message	(guchar *message);
+static void	file_selected		(ImageList *il);
+static void	file_selected_internal	(ImageList *il);
+static void	file_unselected		(ImageList *il);
+static void	dir_entry_enter		(GtkWidget *widget, GtkWidget *entry);
+static void	refresh_status		();
+static void	preview_size_changed	(GtkWidget *box);
+static void	clear_preview_box	();
+static void	auto_refresh_list	();
+static void	show_viewer		();
+static void	print_help_message	();
+static void	print_error_message	(guchar *message);
+
+static void	file_clipboard_clear	();
+static void	file_clipboard_alloc	(guint num);
+static void	file_clipboard_set	(guint num, guchar *file);
+static void	file_paste		(guchar *src_file,
+					 guchar *dest_path,
+					 PasteType type);
+static void	menu_edit_paste_internal(gboolean pastelink);
 
 gboolean
 get_current_image(gchar *buffer)
@@ -129,7 +146,7 @@ refresh_list()
 	set_normal_cursor();
 }
 
-void
+static void
 auto_refresh_list()
 {
 	struct stat st;
@@ -140,22 +157,20 @@ auto_refresh_list()
 	refresh_list();
 }
 
-void
+static void
 clear_preview_box()
 {
-	if (GTK_WIDGET_REALIZED(preview_box))
+	if (GTK_WIDGET_REALIZED(preview_image))
 	{
 		gdk_window_clear_area(
 			preview_image->window,
-			preview_image->allocation.x,
-			preview_image->allocation.y,
+			0, 0,
 			preview_image->allocation.width,
 			preview_image->allocation.height);
 	}
-	gtk_image_set(GTK_IMAGE(preview_image), blank_image, NULL);
 }
 
-void
+static void
 preview_size_changed(GtkWidget *box)
 {
 	guint width, height;
@@ -174,16 +189,18 @@ preview_size_changed(GtkWidget *box)
 	 * after the size has been changed */
 	image_list_remove_cache(IMAGE_LIST(imagelist));
 	if (image_list_get_selected(IMAGE_LIST(imagelist)) == NULL) return;
+
 	/* call this to reload the image */
 	file_selected_internal(IMAGE_LIST(imagelist));
 }
 
-void
+static void
 dir_entry_enter(GtkWidget *widget, GtkWidget *entry)
 {
 	gchar entry_text[256];
 	GtkAdjustment *vadj;
 	gint pos;
+	struct stat st;
 
 	if (busy_level > 0) return;
 	
@@ -197,6 +214,13 @@ dir_entry_enter(GtkWidget *widget, GtkWidget *entry)
 		strcat(entry_text, gtk_entry_get_text(GTK_ENTRY(entry)));
 	}
 	pos = dirtree_select_dir(DIRTREE(tree), entry_text);
+
+	if ((pos < 0) && (stat(entry_text, &st) == 0))
+	{  /* for automounting -lc- */
+		dirtree_refresh(DIRTREE(tree));
+		pos = dirtree_select_dir(DIRTREE(tree), entry_text);
+	}
+
 	if (pos >= 0)
 	{
 		vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(treewin));
@@ -207,7 +231,7 @@ dir_entry_enter(GtkWidget *widget, GtkWidget *entry)
 	}
 }
 
-void
+static void
 set_busy_cursor()
 {
 	if (busy_level == 0)
@@ -219,7 +243,7 @@ set_busy_cursor()
 	busy_level++;
 }
 
-void
+static void
 set_normal_cursor()
 {
 	busy_level--;
@@ -230,14 +254,14 @@ set_normal_cursor()
 	}
 }
 
-void
+static void
 dir_selected(GtkTree *dt, GtkWidget *child)
 {
 	if (busy_level > 0) return;
 	dir_selected_internal(dt, child);
 }
 
-void
+static void
 dir_selected_internal(GtkTree *dt, GtkWidget *child)
 {	
 	struct stat st;
@@ -255,25 +279,29 @@ dir_selected_internal(GtkTree *dt, GtkWidget *child)
 	list_last_modified = st.st_mtime;
 }
 
-void
+static void
 file_selected(ImageList *il)
 {
 	if (busy_level > 0) return;
 	file_selected_internal(il);
 }
 
-void
+static void
 file_selected_internal(ImageList *il)
 {
 	gchar *text;
 	gchar fullname[256], buffer[256];
 	ImageInfo *info = image_list_get_selected(il);
 	struct tm *time;
-	gint hour;
+	gint hour, i;
 	gint max_width, max_height;
-	gint old_width;
+	gint old_width, t;
 	gboolean fast_preview;
+	GtkPreview *preview = GTK_PREVIEW(preview_image);
+	ImageCache *loaded_cache;
+	guchar *ptr;
 
+	if (info == NULL) return;
 	clear_preview_box();
 	
 	toolbar_remove_enable(TRUE);
@@ -285,11 +313,24 @@ file_selected_internal(ImageList *il)
 	/* setting status_file(size, date/time) */
 	text = fsize(info->size);
 	time = localtime(&info->time);
-	hour = time->tm_hour % 12;
-	hour = (hour==0 ? 12 : hour);
-	sprintf(buffer, "%s, %02i/%02i/%02i %02i:%02i %s",
-		text, time->tm_year, time->tm_mon+1, time->tm_mday,
-		hour, time->tm_min, (time->tm_hour<12)?"AM":"PM");
+	if (time->tm_hour <=12)
+	{
+		hour = time->tm_hour;
+	} else
+	{
+		hour = time->tm_hour - 12;
+	}
+	if (hour == 0)
+	{
+		sprintf(buffer, "%s, %02i/%02i/%02i 12:%02i AM",
+			text, time->tm_year, time->tm_mon+1, time->tm_mday,
+			time->tm_min);
+	} else
+	{
+		sprintf(buffer, "%s, %02i/%02i/%02i %02i:%02i %s",
+			text, time->tm_year, time->tm_mon+1, time->tm_mday,
+			hour, time->tm_min, (time->tm_hour<12)?"AM":"PM");
+	}
 	g_free(text);
 	set_status_file(buffer);
 	
@@ -316,20 +357,27 @@ file_selected_internal(ImageList *il)
 	if (info -> valid)
 	{
 		/* load the image if necessary */
-		if (! info -> loaded)
+		if (!info->loaded)
 		{
 			set_busy_cursor();
-			info -> loaded = TRUE;
+			info->loaded = TRUE;
 			strcpy(fullname, image_list_get_dir(il));
 			if (fullname[strlen(fullname) - 1] != '/') strcat(fullname,"/");
 			strcat(fullname, info->name);
 			old_width = info->width;
-			info->image = load_scaled_image(fullname, info,
+			load_scaled_image(fullname, info,
 				max_width, max_height,
-				ROTATE_NONE,
-				il -> parent,
+				il->parent,
 				preview_image,
-				SCANLINE_DISPLAY | (fast_preview?SCANLINE_PREVIEW:0));
+				SCANLINE_DISPLAY |
+				(fast_preview?SCANLINE_PREVIEW:0));
+			loaded_cache = scanline_get_cache();
+			info->cache.buffer_width = loaded_cache->buffer_width;
+			info->cache.buffer_height = loaded_cache->buffer_height;
+			t = loaded_cache->buffer_width *
+				loaded_cache->buffer_height * 3;
+			info->cache.buffer = g_malloc(t);
+			memcpy(info->cache.buffer, loaded_cache->buffer, t);
 			set_normal_cursor();
 			if (info->width != old_width)
 			{
@@ -348,13 +396,23 @@ file_selected_internal(ImageList *il)
 			}
 		} else
 		{
-			/* the image has been load before. just show it. */
-			gtk_image_set(GTK_IMAGE(preview_image), info->image, NULL);
+			/* The image has been load before. Use cache. */
+			gtk_preview_size(preview, info->cache.buffer_width,
+				info->cache.buffer_height);
+			gtk_widget_queue_resize(preview_image);
+			ptr = info->cache.buffer;
+			for (i = 0; i < info->cache.buffer_height; i++)
+			{
+				gtk_preview_draw_row(preview, ptr,
+					0, i, info->cache.buffer_width);
+				ptr += info->cache.buffer_width * 3;
+			}
+			gtk_widget_draw(preview_image, NULL);
 		}
 	}
 }
 
-void
+static void
 file_unselected(ImageList *il)
 {
 	if (busy_level > 0) return;
@@ -392,7 +450,7 @@ menu_file_view(GtkWidget *widget, gpointer data)
 	show_viewer();
 }
 
-void
+static void
 show_viewer()
 {
 	if (busy_level > 0) return;
@@ -457,6 +515,120 @@ menu_file_quit(GtkWidget *widget, gpointer data)
 	gtk_main_quit();
 }
 
+void
+menu_edit_copy(GtkWidget *widget, gpointer data)
+{
+	GList *selection;
+	ImageInfo *info;
+	guint counter;
+	char buf[256];
+
+	selection = image_list_get_selection(IMAGE_LIST(imagelist));
+	if (selection == NULL) return;
+	file_clipboard_clear();
+	file_clipboard_alloc(g_list_length(selection));
+	counter = 0;
+	selection = g_list_first(selection);
+	while (selection != NULL)
+	{
+		info = image_list_get_by_serial(IMAGE_LIST(imagelist),
+			(guint)selection->data);
+		sprintf(buf, "%s/%s", IMAGE_LIST(imagelist)->dir, info->name);
+		file_clipboard_set(counter++, buf);
+		selection = g_list_next(selection);
+	}
+	clipboard_moving = FALSE;
+}
+
+void
+menu_edit_cut(GtkWidget *widget, gpointer data)
+{
+	/* same as copy; except the flag clipboard_moving */
+	menu_edit_copy(widget, data);
+	clipboard_moving = TRUE;
+}
+
+void
+menu_edit_paste(GtkWidget *widget, gpointer data)
+{
+	menu_edit_paste_internal(FALSE);
+}
+
+void
+menu_edit_paste_link(GtkWidget *widget, gpointer data)
+{
+	menu_edit_paste_internal(TRUE);
+}
+
+static void
+menu_edit_paste_internal(gboolean pastelink)
+{
+	gint i;
+	guchar *dir = IMAGE_LIST(imagelist)->dir;
+	guchar *file, buf[256];
+
+	if (file_clipboard == NULL || clipboard_size < 1) return;
+
+	/* cannot paste to the same directory! */
+	if (strncmp(file_clipboard[0], dir, strlen(dir)) == 0 &&
+	    file_clipboard[0][strlen(dir)] == '/' &&
+	    strchr(file_clipboard[0] + strlen(dir) + 1, '/') == NULL)
+	{
+		return;
+	}
+
+	for (i = 0; i < clipboard_size; i++)
+	{
+		file_paste(file_clipboard[i], dir,
+			clipboard_moving ? PASTE_MOVING :
+			(pastelink ? PASTE_LINKING : PASTE_COPYING));
+		/*
+		 * If we are moving files, we should fix the clipboard
+		 * so that we can paste again; and next time should be
+		 * copy, not move. clipboard_moving flag should be fixed
+		 * after the loop.
+		 */
+		if (clipboard_moving)
+		{
+			file = strrchr(file_clipboard[i], '/') + 1;
+			sprintf(buf, "%s/%s", dir, file);
+			file_clipboard_set(i, buf);
+		}
+	}
+	clipboard_moving = FALSE;
+	refresh_list();
+}
+
+void
+menu_edit_delete(GtkWidget *widget, gpointer data)
+{
+	GList *selection;
+	ImageInfo *info;
+	char buf[256];
+
+	selection = image_list_get_selection(IMAGE_LIST(imagelist));
+	if (selection == NULL) return;
+	selection = g_list_first(selection);
+	while (selection != NULL)
+	{
+		info = image_list_get_by_serial(IMAGE_LIST(imagelist),
+			(guint)selection->data);
+		sprintf(buf, "rm -f %s/%s",
+			IMAGE_LIST(imagelist)->dir,
+			info->name);
+		system(buf);
+		selection = g_list_next(selection);
+	}
+	file_clipboard_clear();
+	refresh_list();
+}
+
+void
+menu_edit_select_all(GtkWidget *widget, gpointer data)
+{
+	image_list_select_all(IMAGE_LIST(imagelist));
+}
+
 void
 menu_view_show_hidden(GtkWidget *widget, gpointer data)
 {
@@ -511,6 +683,102 @@ menu_view_preview(GtkWidget *widget, gpointer data)
 	rc_save_gtkseerc();
 }
 
+void
+menu_view_sort_by_name(GtkWidget *widget, gpointer data)
+{
+	ImageSortType type;
+
+	type = rc_get_int("image_sort_type");
+	if (type == RC_NOT_EXISTS)
+	{
+		type = IMAGE_SORT_ASCEND_BY_NAME;
+	} else
+	{
+		type = type % 2;
+	}
+	image_list_set_sort_type(IMAGE_LIST(imagelist), type);
+}
+
+void
+menu_view_sort_by_size(GtkWidget *widget, gpointer data)
+{
+	ImageSortType type;
+
+	type = rc_get_int("image_sort_type");
+	if (type == RC_NOT_EXISTS)
+	{
+		type = IMAGE_SORT_ASCEND_BY_SIZE;
+	} else
+	{
+		type = type % 2 + 2;
+	}
+	image_list_set_sort_type(IMAGE_LIST(imagelist), type);
+}
+
+void
+menu_view_sort_by_property(GtkWidget *widget, gpointer data)
+{
+	ImageSortType type;
+
+	type = rc_get_int("image_sort_type");
+	if (type == RC_NOT_EXISTS)
+	{
+		type = IMAGE_SORT_ASCEND_BY_PROPERTY;
+	} else
+	{
+		type = type % 2 + 4;
+	}
+	image_list_set_sort_type(IMAGE_LIST(imagelist), type);
+}
+
+void
+menu_view_sort_by_date(GtkWidget *widget, gpointer data)
+{
+	ImageSortType type;
+
+	type = rc_get_int("image_sort_type");
+	if (type == RC_NOT_EXISTS)
+	{
+		type = IMAGE_SORT_ASCEND_BY_DATE;
+	} else
+	{
+		type = type % 2 + 6;
+	}
+	image_list_set_sort_type(IMAGE_LIST(imagelist), type);
+}
+
+void
+menu_view_sort_ascend(GtkWidget *widget, gpointer data)
+{
+	ImageSortType type;
+
+	type = rc_get_int("image_sort_type");
+	if (type == RC_NOT_EXISTS)
+	{
+		type = IMAGE_SORT_ASCEND_BY_NAME;
+	} else
+	{
+		type = type - type % 2;
+	}
+	image_list_set_sort_type(IMAGE_LIST(imagelist), type);
+}
+
+void
+menu_view_sort_descend(GtkWidget *widget, gpointer data)
+{
+	ImageSortType type;
+
+	type = rc_get_int("image_sort_type");
+	if (type == RC_NOT_EXISTS)
+	{
+		type = IMAGE_SORT_DESCEND_BY_NAME;
+	} else
+	{
+		type = type - type % 2 + 1;
+	}
+	image_list_set_sort_type(IMAGE_LIST(imagelist), type);
+}
+
 void
 menu_view_quick_refresh(GtkWidget *widget, gpointer data)
 {
@@ -640,13 +908,13 @@ toolbar_details(GtkWidget *widget, gpointer data)
 	}
 }
 
-void
+static void
 refresh_status()
 {
 	gchar buffer[64];
 	
 	/* set status_main(total files and size) */
-	sprintf(buffer, "Total %i file(s) (%s)",
+	sprintf(buffer, _("Total %i file(s) (%s)"),
 		IMAGE_LIST(imagelist)->nfiles,
 		fsize(IMAGE_LIST(imagelist)->total_size));
 	set_status_main(buffer);
@@ -664,7 +932,7 @@ refresh_status()
 	toolbar_view_enable(image_list_get_first(IMAGE_LIST(imagelist)) != NULL);
 }
 
-void
+static void
 print_help_message()
 {
 	printf("Usage: gtksee [-R[directory]] [-fis] [files...]\n");
@@ -678,7 +946,7 @@ print_help_message()
 	_exit(0);
 }
 
-void
+static void
 print_error_message(guchar *message)
 {
 	printf("Error: %s\n", message);
@@ -694,8 +962,15 @@ main (int argc, char *argv[])
 	ImageInfo info, *pinfo;
 	char *j;
 	guchar *r;
+
+#ifdef ENABLE_NLS
+	bindtextdomain(PACKAGE, LOCALEDIR);
+	textdomain(PACKAGE);
+#endif
+
+	gtk_set_locale();
 	
-	gtk_init (&argc, &argv);
+	gtk_init(&argc, &argv);
 
 	rc_init();
 
@@ -781,7 +1056,7 @@ main (int argc, char *argv[])
 	gtk_main();
 	return 0;
 }
-	
+
 void
 gtksee_main()
 {
@@ -863,11 +1138,8 @@ gtksee_main()
 		GTK_SIGNAL_FUNC(preview_size_changed),
 		GTK_OBJECT(preview_box));
 	gtk_widget_show(preview_box);
-	
-	blank_image = gdk_image_new(GDK_IMAGE_NORMAL,
-		gdk_window_get_visual(window->window),
-		0, 0);
-	preview_image = gtk_image_new(blank_image, NULL);
+
+	preview_image = gtk_preview_new(GTK_PREVIEW_COLOR);
 	gtk_container_add(GTK_CONTAINER(preview_box), preview_image);
 	gtk_widget_show(preview_image);
 	
@@ -983,3 +1255,64 @@ gtksee_main()
 	}
 
 }
+
+static void
+file_clipboard_clear()
+{
+	gint i;
+
+	if (file_clipboard == NULL) return;
+	for (i = 0; i < clipboard_size; i++)
+	{
+		if (file_clipboard[i] != NULL) free(file_clipboard[i]);
+	}
+	free(file_clipboard);
+	file_clipboard = NULL;
+	clipboard_size = 0;
+}
+
+static void
+file_clipboard_alloc(guint num)
+{
+	gint i;
+
+	clipboard_size = num;
+	if (num < 1) return;
+	file_clipboard = g_malloc(sizeof(guchar*) * num);
+	for (i = 0; i < num; i++)
+	{
+		file_clipboard[i] = NULL;
+	}
+}
+
+static void
+file_clipboard_set(guint num, guchar *file)
+{
+	if (file_clipboard[num] != NULL) free(file_clipboard[num]);
+	file_clipboard[num] = g_malloc(sizeof(guchar) * strlen(file) + 1);
+	strcpy(file_clipboard[num], file);
+}
+
+static void
+file_paste(guchar *src_file, guchar *dest_path, PasteType type)
+{
+	char buf[512];
+
+	switch(type)
+	{
+	  case PASTE_MOVING:
+		sprintf(buf, "mv -f %s %s", src_file, dest_path);
+		system(buf);
+		break;
+	  case PASTE_COPYING:
+		sprintf(buf, "cp -rf %s %s", src_file, dest_path);
+		system(buf);
+		break;
+	  case PASTE_LINKING:
+		sprintf(buf, "%s/%s", dest_path, strrchr(src_file, '/') + 1);
+		link(src_file, buf);
+		break;
+	  default:
+		break;
+	}
+}
