@@ -30,20 +30,20 @@
 #define SCANLINE_INTERACT 1
 #define SCANLINE_DISPLAY 2
 #define SCANLINE_PREVIEW 4
-#define SCANLINE_CONTAINER 8
 
 void		scanline_init	(GdkColorContext *gcc, GtkStyle *style);
 void		scanline_start	();
 void		scanline_cancel	();
-GdkImage*	load_scaled_image
+void		load_scaled_image
 				(gchar *orig_filename,
 				 ImageInfo *info,
 				 gint max_width,
 				 gint max_height,
-				 RotateType r_type,
 				 GtkWidget *window,
 				 GtkWidget *container,
 				 gint scanflags);
+ImageCache*	scanline_get_cache
+				(void);
 
 #ifdef __cplusplus
 	}
