@@ -17,12 +17,15 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
+#include "intl.h"
 #include "gtypes.h"
 #include "scanline.h"
 #include "imagelist.h"
@@ -45,17 +48,16 @@
 
 static GtkWidget *window, *imagelist;
 static GList *image_infos;
-static GdkImage *loaded_image;
 static ImageInfo *current;
 static GtkWidget *frame;
-static RotateType rtype;
 static gchar viewing_image[256];
 static gboolean viewer_busy;
 static GdkCursor *viewer_watch_cursor = NULL;
-static GtkWidget *viewport;
+static GtkWidget *viewport, *viewport_image;
 
 static gboolean slide_show = FALSE;
 static guint timeout_callback_tag = 0;
+static gboolean have_xgrab;
 
 void		view_current_image		();
 void		viewer_set_busy_cursor		();
@@ -108,11 +110,8 @@ void
 view_current_image()
 {
 	gboolean too_big, need_scale;
-	GList *list;
 	gchar buffer[256], *ptr;
 	gint width, height, max_width, max_height, tmp;
-	GtkStyle *style;
-	GdkImage *new_image;
 	gint value;
 	gboolean full_screen, fit_screen;
 	
@@ -146,7 +145,7 @@ view_current_image()
 			}
 			set_viewer_status_name(ptr);
 			set_viewer_status_size(current->size);
-			set_viewer_status_prop("Invalid");
+			set_viewer_status_prop(_("Invalid"));
 			set_viewer_status_zoom(100);
 		}
 		viewer_set_normal_cursor();
@@ -199,41 +198,11 @@ view_current_image()
 	max_width = (full_screen?MAX_FULL_WIDTH:MAX_WIDTH);
 	max_height = (full_screen?MAX_FULL_HEIGHT:MAX_HEIGHT);
 	
-	switch (rtype)
-	{
-		case ROTATE_90:
-		case ROTATE_270:
-			width = current->height;
-			height = current->width;
-			break;
-		case ROTATE_180:
-		case ROTATE_NONE:
-		default:
-			width = current->width;
-			height = current->height;
-			break;
-	}
+	width = current->width;
+	height = current->height;
 	too_big = (width > max_width || height > max_height);
 	need_scale = (fit_screen && too_big);
 	
-	viewport = dnd_viewport_new();
-	gtk_signal_connect(
-		GTK_OBJECT(viewport),
-		"double_clicked",
-		GTK_SIGNAL_FUNC(viewer_toolbar_browse),
-		NULL);
-	if (full_screen)
-	{
-		style = gtk_style_new();
-		style->bg[GTK_STATE_NORMAL] = style->black;
-		gtk_widget_set_style(viewport, style);
-		/* use popup menu only in full-screen mode */
-		gtk_signal_connect(
-			GTK_OBJECT(viewport),
-			"popup",
-			GTK_SIGNAL_FUNC(viewer_menu_popup),
-			NULL);
-	}
 	if (need_scale)
 	{
 		if (max_width * height < max_height * width)
@@ -271,68 +240,40 @@ view_current_image()
 			min(height,max_height));
 	}
 
-	/* In slide-show mode, we should replace new image after loaded */
-	if (!slide_show)
-	{
-		/* remove old image if exists */
-		list = gtk_container_children(GTK_CONTAINER(frame));
-		if (list != NULL && list->data != NULL)
-		{
-			gtk_container_remove(
-				GTK_CONTAINER(frame),
-				GTK_WIDGET(list->data));
-		}
-		if (loaded_image != NULL) gdk_image_destroy(loaded_image);
-		loaded_image = NULL;
-		gtk_container_add(GTK_CONTAINER(frame), viewport);
-		gtk_widget_show(viewport);
-	}
-	
+	/* clear last image */
+	gdk_window_clear_area(
+		viewport_image->window,
+		0, 0,
+		viewport_image->allocation.width,
+		viewport_image->allocation.height);
+
 	/* allow the user to drag the viewport while displaying */
 	gtk_grab_add(viewport);
-	
+
+	/* reset the viewport position to (0,0) */
+	dnd_viewport_reset(DND_VIEWPORT(viewport));
+
 	/* for slideshow: disable timeout temporarily before loading */
 	if (slide_show && timeout_callback_tag != 0)
 	{
 		gtk_timeout_remove(timeout_callback_tag);
 	}
 	
-	/* creating new image */
 	if (need_scale)
 	{
-		new_image = load_scaled_image(viewing_image, current,
+		load_scaled_image(viewing_image, current,
 			max_width, max_height,
-			rtype,
 			window,
-			viewport,
-			SCANLINE_CONTAINER |
+			viewport_image,
 			(slide_show ? 0 : (SCANLINE_INTERACT | SCANLINE_DISPLAY)));
 	} else
 	{
-		new_image = load_scaled_image(viewing_image, current,
+		load_scaled_image(viewing_image, current,
 			-1, -1,
-			rtype,
 			window,
-			viewport,
-			SCANLINE_CONTAINER |
+			viewport_image,
 			(slide_show ? 0 : (SCANLINE_INTERACT | SCANLINE_DISPLAY)));
 	}
-	if (slide_show)
-	{
-		/* remove old image if exists */
-		list = gtk_container_children(GTK_CONTAINER(frame));
-		if (list != NULL && list->data != NULL)
-		{
-			gtk_container_remove(
-				GTK_CONTAINER(frame),
-				GTK_WIDGET(list->data));
-		}
-		if (loaded_image != NULL) gdk_image_destroy(loaded_image);
-		loaded_image = NULL;
-		gtk_container_add(GTK_CONTAINER(frame), viewport);
-		gtk_widget_show(viewport);
-	}
-	loaded_image = new_image;
 	gtk_grab_remove(viewport);
 	if (slide_show)
 	{
@@ -349,9 +290,12 @@ viewer_toolbar_browse(GtkWidget *widget, gpointer data)
 {
 	if (viewer_busy) return;
 	
-	if (loaded_image != NULL) gdk_image_destroy(loaded_image);
-	loaded_image = NULL;
 	if (slide_show) gtk_timeout_remove(timeout_callback_tag);
+	if (have_xgrab)
+	{
+		gdk_pointer_ungrab(0);
+		gdk_keyboard_ungrab(0);
+	}
 	gtk_widget_destroy(window);
 	if (imagelist == NULL)
 	{
@@ -370,8 +314,6 @@ viewer_toolbar_full_screen(GtkWidget *widget, gpointer data)
 	
 	rc_set_boolean("full_screen", TRUE);
 	rc_save_gtkseerc();
-	if (loaded_image != NULL) gdk_image_destroy(loaded_image);
-	loaded_image = NULL;
 	gtk_widget_destroy(window);
 	get_full_view_window();
 }
@@ -411,7 +353,6 @@ viewer_toolbar_next_image(GtkWidget *widget, gpointer data)
 			viewer_prev_enable(image_list_get_previous(IMAGE_LIST(imagelist), current) != NULL);
 		}
 	}
-	rtype = ROTATE_NONE;
 	view_current_image();
 }
 
@@ -450,7 +391,6 @@ viewer_toolbar_prev_image(GtkWidget *widget, gpointer data)
 			viewer_prev_enable(image_list_get_previous(IMAGE_LIST(imagelist), current) != NULL);
 		}
 	}
-	rtype = ROTATE_NONE;
 	view_current_image();
 }
 
@@ -535,26 +475,101 @@ viewer_slideshow_next(gpointer data)
 			viewer_prev_enable(image_list_get_previous(IMAGE_LIST(imagelist), current) != NULL);
 		}
 	}
-	rtype = ROTATE_NONE;
 	view_current_image();
 }
 
 void
 viewer_toolbar_rotate_left(GtkWidget *widget, gpointer data)
 {
+	ImageCache *cache;
+	gint i, j, bufsize, t;
+	guchar *newbuf, *ptr1, *ptr2;
+
 	if (viewer_busy) return;
 
-	rtype = (rtype - 1 + MAX_ROTATE_TYPES) % MAX_ROTATE_TYPES;
-	view_current_image();
+	cache = scanline_get_cache();
+	bufsize = cache->buffer_width * cache->buffer_height * 3;
+	newbuf = g_malloc(bufsize);
+	for (i = 0; i < cache->buffer_height; i++)
+	{
+		for (j = 0; j < cache->buffer_width; j++)
+		{
+			ptr1 = newbuf + ((cache->buffer_width - j - 1)
+				* cache->buffer_height + i) * 3;
+			ptr2 = cache->buffer +
+				(i * cache->buffer_width + j) * 3;
+			*ptr1++ = *ptr2++;
+			*ptr1++ = *ptr2++;
+			*ptr1++ = *ptr2++;
+		}
+	}
+	g_free(cache->buffer);
+	cache->buffer = newbuf;
+	t = cache->buffer_width;
+	cache->buffer_width = cache->buffer_height;
+	cache->buffer_height = t;
+
+	/* Display the rotated image */
+	gtk_preview_size(GTK_PREVIEW(viewport_image),
+		cache->buffer_width, cache->buffer_height);
+	gtk_widget_queue_resize(viewport_image);
+	for (i = 0; i < cache->buffer_height; i++)
+	{
+		gtk_preview_draw_row(GTK_PREVIEW(viewport_image),
+			cache->buffer + i * cache->buffer_width * 3,
+			0, i, cache->buffer_width);
+	}
+	gtk_widget_set_usize(viewport,
+		min(cache->buffer_width, MAX_WIDTH),
+		min(cache->buffer_height, MAX_HEIGHT));
+	gtk_widget_draw(viewport_image, NULL);
 }
 
 void
 viewer_toolbar_rotate_right(GtkWidget *widget, gpointer data)
 {
+	ImageCache *cache;
+	gint i, j, bufsize, t;
+	guchar *newbuf, *ptr1, *ptr2;
+
 	if (viewer_busy) return;
 
-	rtype = (rtype + 1) % MAX_ROTATE_TYPES;
-	view_current_image();
+	cache = scanline_get_cache();
+	bufsize = cache->buffer_width * cache->buffer_height * 3;
+	newbuf = g_malloc(bufsize);
+	for (i = 0; i < cache->buffer_height; i++)
+	{
+		for (j = 0; j < cache->buffer_width; j++)
+		{
+			ptr1 = newbuf + (j * cache->buffer_height +
+				(cache->buffer_height - i - 1)) * 3;
+			ptr2 = cache->buffer +
+				(i * cache->buffer_width + j) * 3;
+			*ptr1++ = *ptr2++;
+			*ptr1++ = *ptr2++;
+			*ptr1++ = *ptr2++;
+		}
+	}
+	g_free(cache->buffer);
+	cache->buffer = newbuf;
+	t = cache->buffer_width;
+	cache->buffer_width = cache->buffer_height;
+	cache->buffer_height = t;
+
+	/* Display the rotated image */
+	gtk_preview_size(GTK_PREVIEW(viewport_image),
+		cache->buffer_width, cache->buffer_height);
+	gtk_widget_queue_resize(viewport_image);
+	for (i = 0; i < cache->buffer_height; i++)
+	{
+		gtk_preview_draw_row(GTK_PREVIEW(viewport_image),
+			cache->buffer + i * cache->buffer_width * 3,
+			0, i, cache->buffer_width);
+	}
+	gtk_widget_set_usize(viewport,
+		min(cache->buffer_width, MAX_WIDTH),
+		min(cache->buffer_height, MAX_HEIGHT));
+	gtk_widget_draw(viewport_image, NULL);
 }
 
 void
@@ -566,23 +581,52 @@ viewer_toolbar_refresh(GtkWidget *widget, gpointer data)
 void
 viewer_key_release(GtkWidget *widget, GdkEvent *event, gpointer data)
 {
+	GList *tmplist;
+	ImageInfo *tmp;
+	gboolean full_screen;
+
 	if (viewer_busy) return;
 	
+	full_screen = rc_get_boolean("full_screen");
+
 	switch (event->key.keyval)
 	{
+	  case ' ':
+		if (imagelist == NULL)
+		{
+			tmplist = g_list_next(image_infos);
+			if (tmplist == NULL) break;
+			image_infos = tmplist;
+			tmp = (ImageInfo*) image_infos->data;
+		} else
+		{
+			tmp = image_list_get_next(IMAGE_LIST(imagelist), current);
+		}
+		if (tmp != NULL)
+		{
+			viewer_toolbar_next_image(widget, NULL);
+			break;
+		}
+		/* Loop to the first if we are at the end of the list */
 	  case GDK_Home:
 		if (imagelist == NULL)
 		{
 			image_infos = g_list_first(image_infos);
 			current = (ImageInfo*) image_infos->data;
-			viewer_next_enable(g_list_next(image_infos) != NULL);
+			if (!full_screen)
+			{
+				viewer_next_enable(g_list_next(image_infos) != NULL);
+				viewer_prev_enable(FALSE);
+			}
 		} else
 		{
 			current = image_list_get_first(IMAGE_LIST(imagelist));
-			viewer_next_enable(image_list_get_next(IMAGE_LIST(imagelist), current) != NULL);
+			if (!full_screen)
+			{
+				viewer_next_enable(image_list_get_next(IMAGE_LIST(imagelist), current) != NULL);
+				viewer_prev_enable(FALSE);
+			}
 		}
-		viewer_prev_enable(FALSE);
-		rtype = ROTATE_NONE;
 		view_current_image();
 		break;
 	  case GDK_End:
@@ -590,21 +634,26 @@ viewer_key_release(GtkWidget *widget, GdkEvent *event, gpointer data)
 		{
 			image_infos = g_list_last(image_infos);
 			current = (ImageInfo*) image_infos->data;
-			viewer_prev_enable(g_list_previous(image_infos) != NULL);
+			if (!full_screen)
+			{
+				viewer_prev_enable(g_list_previous(image_infos) != NULL);
+				viewer_next_enable(FALSE);
+			}
 		} else
 		{
 			current = image_list_get_last(IMAGE_LIST(imagelist));
-			viewer_prev_enable(image_list_get_previous(IMAGE_LIST(imagelist), current) != NULL);
+			if (!full_screen)
+			{
+				viewer_prev_enable(image_list_get_previous(IMAGE_LIST(imagelist), current) != NULL);
+				viewer_next_enable(FALSE);
+			}
 		}
-		viewer_next_enable(FALSE);
-		rtype = ROTATE_NONE;
 		view_current_image();
 		break;
 	  case GDK_Page_Up:
 		viewer_toolbar_prev_image(widget, NULL);
 		break;
 	  case GDK_Page_Down:
-	  case ' ':
 		viewer_toolbar_next_image(widget, NULL);
 		break;
 	  case GDK_Escape:
@@ -641,7 +690,7 @@ viewer_menu_popup(GtkWidget *widget, gpointer data)
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_menu_item_new_with_label("Next image");
+	menu_item = gtk_menu_item_new_with_label(_("Next image"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
 		GTK_SIGNAL_FUNC(viewer_toolbar_next_image), NULL);
@@ -658,7 +707,7 @@ viewer_menu_popup(GtkWidget *widget, gpointer data)
 	}
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_menu_item_new_with_label("Previous image");
+	menu_item = gtk_menu_item_new_with_label(_("Previous image"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
 		GTK_SIGNAL_FUNC(viewer_toolbar_prev_image), NULL);
@@ -675,7 +724,7 @@ viewer_menu_popup(GtkWidget *widget, gpointer data)
 	}
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_check_menu_item_new_with_label("Slideshow");
+	menu_item = gtk_check_menu_item_new_with_label(_("Slide show"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	if (slide_show)
 	{
@@ -698,7 +747,7 @@ viewer_menu_popup(GtkWidget *widget, gpointer data)
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_check_menu_item_new_with_label("Fit screen");
+	menu_item = gtk_check_menu_item_new_with_label(_("Fit screen"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 
 	fit_screen = rc_get_boolean("fit_screen");
@@ -719,7 +768,7 @@ viewer_menu_popup(GtkWidget *widget, gpointer data)
 	}
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_check_menu_item_new_with_label("Full screen");
+	menu_item = gtk_check_menu_item_new_with_label(_("Full screen"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 
 	full_screen = rc_get_boolean("full_screen");
@@ -744,19 +793,19 @@ viewer_menu_popup(GtkWidget *widget, gpointer data)
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_menu_item_new_with_label("Browser");
+	menu_item = gtk_menu_item_new_with_label(_("Browser"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
 		GTK_SIGNAL_FUNC(viewer_toolbar_browse), NULL);
 	gtk_widget_show(menu_item);
 	
-	menu_item = gtk_menu_item_new_with_label("Exit");
+	menu_item = gtk_menu_item_new_with_label(_("Exit"));
 	gtk_menu_append(GTK_MENU(menu), menu_item);
 	gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
 		GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
 	gtk_widget_show(menu_item);
 	
-	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 3, 0);
+	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 3, 1);
 }
 
 void
@@ -766,8 +815,6 @@ viewer_popup_view(GtkWidget *widget, gpointer data)
 	
 	rc_set_boolean("full_screen", FALSE);
 	rc_save_gtkseerc();
-	if (loaded_image != NULL) gdk_image_destroy(loaded_image);
-	loaded_image = NULL;
 	gtk_widget_destroy(window);
 	get_viewer_window_internal();
 }
@@ -798,13 +845,11 @@ get_viewer_window(GtkWidget *il)
 	/* setting pointers and values... */
 	imagelist = il;
 	image_infos = NULL;
-	rtype = ROTATE_NONE;
 	if (viewer_watch_cursor == NULL)
 	{
 		viewer_watch_cursor = gdk_cursor_new(GDK_WATCH);
 	}
 	
-	loaded_image = NULL;
 	viewer_busy = FALSE;
 	
 	/* view the selected image if selected,
@@ -815,6 +860,7 @@ get_viewer_window(GtkWidget *il)
 		current = image_list_get_first(IMAGE_LIST(imagelist));
 	
 	full_screen = rc_get_boolean("full_screen");
+	have_xgrab = FALSE;
 	if (full_screen)
 	{
 		return get_full_view_window();
@@ -832,13 +878,11 @@ get_viewer_window_with_files(GList *infos)
 	/* setting pointers and values... */
 	imagelist = NULL;
 	image_infos = g_list_first(infos);
-	rtype = ROTATE_NONE;
 	if (viewer_watch_cursor == NULL)
 	{
 		viewer_watch_cursor = gdk_cursor_new(GDK_WATCH);
 	}
 	
-	loaded_image = NULL;
 	viewer_busy = FALSE;
 	
 	/* view the selected image if selected,
@@ -849,6 +893,7 @@ get_viewer_window_with_files(GList *infos)
 		gtksee_main();
 	
 	full_screen = rc_get_boolean("full_screen");
+	have_xgrab = FALSE;
 	if (full_screen)
 	{
 		return get_full_view_window();
@@ -877,14 +922,20 @@ get_viewer_window_internal()
 	gtk_window_set_title(GTK_WINDOW(window), "GTK See");
 	gtk_window_set_wmclass(GTK_WINDOW(window), "viewer", "GTKSee");
 	event_mask = gtk_widget_get_events(window);
-	event_mask |= GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK;
+	event_mask |= GDK_KEY_RELEASE_MASK;
 	gtk_widget_set_events(window, event_mask);
 	gtk_signal_connect(
 		GTK_OBJECT(window),
 		"key_release_event",
 		GTK_SIGNAL_FUNC(viewer_key_release),
 		NULL);
+#if ((GTK_MAJOR_VERSION>1) || \
+     (GTK_MAJOR_VERSION==1 && GTK_MINOR_VERSION>2) || \
+     (GTK_MAJOR_VERSION==1 && GTK_MINOR_VERSION==2 && GTK_MICRO_VERSION>=5))
+	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ALWAYS);
+#else
 	gtk_widget_set_uposition(window, 0, 0);
+#endif
 	
 	gtk_widget_realize(window);
 	
@@ -910,6 +961,19 @@ get_viewer_window_internal()
 	frame = gtk_frame_new(NULL);
 	gtk_widget_show(frame);
 	
+	viewport = dnd_viewport_new();
+	gtk_signal_connect(
+		GTK_OBJECT(viewport),
+		"double_clicked",
+		GTK_SIGNAL_FUNC(viewer_toolbar_browse),
+		NULL);
+	gtk_container_add(GTK_CONTAINER(frame), viewport);
+	gtk_widget_show(viewport);
+
+	viewport_image = gtk_preview_new(GTK_PREVIEW_COLOR);
+	gtk_container_add(GTK_CONTAINER(viewport), viewport_image);
+	gtk_widget_show(viewport_image);
+
 	/* creating viewer statusbar */
 	pixmap = gdk_pixmap_create_from_xpm_d(
 		window->window,
@@ -928,7 +992,6 @@ get_viewer_window_internal()
 	gtk_table_attach(GTK_TABLE(table), statusbar, 0, 1, 2, 3, GTK_FILL|GTK_EXPAND, 0, 0, 2);
 	
 	/* adding table into viewer window */
-	gtk_widget_unrealize(window);
 	gtk_container_add(GTK_CONTAINER(window), table);
 	gtk_widget_show(window);
 	
@@ -958,14 +1021,13 @@ get_full_view_window()
 	GtkStyle *style;
 	
 	window = gtk_window_new(GTK_WINDOW_POPUP);
+
 	gtk_container_border_width(GTK_CONTAINER(window), 0);
-	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
-		GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
 	gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
 	gtk_widget_set_usize(window, MAX_FULL_WIDTH, MAX_FULL_HEIGHT);
 	gtk_widget_set_uposition(window, 0, 0);
 	event_mask = gtk_widget_get_events(window);
-	event_mask |= GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK;
+	event_mask |= GDK_KEY_RELEASE_MASK;
 	gtk_widget_set_events(window, event_mask);
 	gtk_signal_connect(
 		GTK_OBJECT(window),
@@ -979,7 +1041,32 @@ get_full_view_window()
 	/* cheating frame */
 	frame = window;
 	
+	viewport = dnd_viewport_new();
+	gtk_signal_connect(
+		GTK_OBJECT(viewport),
+		"double_clicked",
+		GTK_SIGNAL_FUNC(viewer_toolbar_browse),
+		NULL);
+	gtk_widget_set_style(viewport, style);
+	/* use popup menu only in full-screen mode */
+	gtk_signal_connect(
+		GTK_OBJECT(viewport),
+		"popup",
+		GTK_SIGNAL_FUNC(viewer_menu_popup),
+		NULL);
+	gtk_container_add(GTK_CONTAINER(frame), viewport);
+	gtk_widget_show(viewport);
+
+	viewport_image = gtk_preview_new(GTK_PREVIEW_COLOR);
+	gtk_widget_set_style(viewport_image, style);
+	gtk_container_add(GTK_CONTAINER(viewport), viewport_image);
+	gtk_widget_show(viewport_image);
+
+	gtk_widget_realize(window);
+	gdk_window_set_decorations(window->window, 0);
+	gdk_window_set_functions(window->window, 0);
 	gtk_widget_show(window);
+
 	if (imagelist == NULL)
 	{
 		image_list_load_pixmaps(window);
@@ -990,5 +1077,22 @@ get_full_view_window()
 	}
 	
 	view_current_image();
+
+	/* Make X grab to full-screen window */
+	if ((gdk_pointer_grab(window->window, TRUE,
+		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
+		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
+		GDK_POINTER_MOTION_MASK,
+		NULL, NULL, 0) == 0))
+	{
+		if (gdk_keyboard_grab(window->window, TRUE, 0) != 0)
+		{
+			gdk_pointer_ungrab(0);
+		} else
+		{
+			have_xgrab = TRUE;
+		}
+	}
+
 	return window;
 }
